<?php

use TYPO3\CMS\Core\Utility\GeneralUtility;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2011-2019 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * The class tx_browser_pi1_download bundles methods for downloading datas
 *
 * @author    Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package    TYPO3
 * @subpackage  browser
 *
 * @version 8.0.12
 * @since 3.9.3
 */
class tx_browser_pi1_download
{

	// [Integer] Number of the current typeNum
	var $int_typeNum = null;
	// [String] Name of the current typeNum
	var $str_typeNum = null;
	// [Boolean] Is dwonloading allowed?
	var $bool_downloadsAllowed = false;
	// [String] view: list || single
	var $view = null;
	// [Integer] mode (index) of the current view
	var $mode = null;
	// [String] table: label of the current table
	var $table = null;
	// [Integer] uid: uid of the current record
	var $uid = null;
	// [String] field: label of the field with the files
	var $field = null;
	// [Integer] key: index of the file
	var $key = null;
	private $typo3Version = null;

	// [INTEGER] Current TYPO3 version as an integer like 4007007 for 4.7.7

	/**
	 * Constructor. The method initiate the parent object
	 *
	 * @param	object		The parent object
	 * @return	void
	 */
	function __construct( $pObj )
	{
		$this->pObj = $pObj;
	}

	/**
	 * caseDAM( ) : 
	 * 
	 * @return	boolean
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function caseDAM()
	{
		if( $this->table == 'tx_dam' )
		{
			return TRUE;
		}

		return FALSE;
	}

	/**
	 * caseDefault( ) : 
	 * 
	 * @return	boolean
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function caseDefault()
	{
		$uploadFolder = $GLOBALS[ 'TCA' ][ $this->table ][ 'columns' ][ $this->field ][ 'config' ][ 'uploadfolder' ];

		if( $uploadFolder )
		{
			return TRUE;
		}

		return FALSE;
	}

	/**
	 * caseFAL( ) : 
	 * 
	 * @return	boolean
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function caseFAL()
	{
		$foreign_table = $GLOBALS[ 'TCA' ][ $this->table ][ 'columns' ][ $this->field ][ 'config' ][ 'foreign_table' ];

		if( $foreign_table == 'sys_file_reference' )
		{
			return TRUE;
		}

		return FALSE;
	}

	/*	 * *********************************************
	 *
	 * Main
	 *
	 * ******************************************** */

	/**
	 * download( ): Main method for downloading the requested file.
	 *              If there is an failure, the method retirns a failure prompt.
	 *              If there is success, this class will send the header, the file
	 *              and will exit the PHP script.
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 3.9.3
	 * @since 3.9.3
	 */
	public function download()
	{
//////////////////////////////////////////////////////////////////////////
//
		// RETURN typeNum isn't the download typeNum

		if( $this->str_typeNum != 'download' )
		{
			return;
		}
// RETURN typeNum isn't the download typeNum
//////////////////////////////////////////////////////////////////////////
//
		// Init global class vars

		$prompt_error = $this->download_init();
		if( $prompt_error )
		{
			$prompt_error = '' .
							'<div style="border:1em solid red;text-align:center;padding:1em;"><h1>TYPO3 Browser</h1>' .
							$prompt_error .
							'</div>';
			return $prompt_error;
		}
// Init global class vars
//////////////////////////////////////////////////////////////////////////
//
		// Check table, field and TypoScript

		$prompt_error = $this->download_check();
		if( $prompt_error )
		{
			$prompt_error = '' .
							'<div style="border:1em solid red;text-align:center;padding:1em;"><h1>TYPO3 Browser</h1>' . $prompt_error . '</div>';
			return $prompt_error;
		}
// Check table, field and TypoScript




		$this->statistics( '+' );
// EXIT in case of success!
		$prompt_error = $this->sendFileAndExit();

		$this->statistics( '-' );

		$prompt_error = '' .
						'<div style="border:1em solid red;text-align:center;padding:1em;"><h1>TYPO3 Browser</h1>' . $prompt_error . '</div>';
		return $prompt_error;
	}

	/**
	 * download_check( ): The method checks
	 *                    * view and mode has to exist
	 *                    * table.field has to be a part of the TypoScript select property
	 *                    * table.field has to be configured in the TCA
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 7.2.16
	 * @since 3.9.3
	 */
	private function download_check()
	{
// Does the view and the mode exist?
		if( !isset( $this->pObj->conf[ 'views.' ][ $this->view . '.' ][ $this->mode . '.' ][ 'select' ] ) )
		{
// #i0204, 151130, dwildt, ~
			$prompt = ''
							. '<p>'
							. '  Security check: TypoScript property '
							. '  plugin.tx_browser_pi1.views.' . $this->view . '. ' . $this->mode . '.select doesn\'t exist.<br />'
							. '</p>'
							. '<h2>'
							. '  Possible causes '
							. '</h2>'
							. '<ul>'
							. '  <li>'
							. '    Organiser pages has an unproper uid. See: Constant Editor [ORGANISER - PAGES]'
							. '  </li>'
							. '  <li>'
							. '    You are using more than one Browser plugin on the current page. And the current plugin ignores piVars from other plugins.<br />'
							. '    Please check the configuration of the current plugin > [General] > Views: piVars (parameters in the URL) from foreign plugins: [Handle it!] '
							. '  </li>'
							. '</ul>'
							. '<h2>'
							. '  Values '
							. '</h2>'
							. '<ul>'
							. '  <li>'
							. '    view: "' . $this->view . '" <- should be single'
							. '  </li>'
							. '  <li>'
							. '    mode: "' . $this->view . '" <- must be the uid of the current single view, which is the base for the download'
							. '  </li>'
							. '</ul>'
							. __METHOD__ . ' (' . __LINE__ . ')';
			return $prompt;
		}

// Is table.field part of the select?
//var_dump( __METHOD__, __LINE__, $this->view, $this->mode, $this->pObj->conf['views.'][$this->view . '.'][$this->mode . '.']['select']);
		$select = $this->pObj->conf[ 'views.' ][ $this->view . '.' ][ $this->mode . '.' ][ 'select' ];
		$pos = strpos( $select, $this->table . '.' . $this->field );
		if( $pos === false )
		{
			$prompt = '' .
							'Security check: ' . $this->table . '.' . $this->field . ' ' .
							'isn\'t part of plugin.tx_browser_pi1.views.' . $this->view . '.' . $this->mode . '.select.<br />' .
							__METHOD__ . ' (' . __LINE__ . ')';
			return $prompt;
		}

// Is table.field part of the TCA? Is field a file type?
// Load the TCA for the current table
// #t0606, 160701, dwildt, 1-
//$this->pObj->objZz->loadTCA($table);
// Check, if the field is an element of the current table
		if( !isset( $GLOBALS[ 'TCA' ][ $this->table ][ 'columns' ][ $this->field ] ) )
		{
			$prompt = '' .
							'Security check: ' . $this->table . '.' . $this->field . ' ' .
							'isn\'t part of the TCA.<br />' .
							__METHOD__ . ' (' . __LINE__ . ')';
			return $prompt;
		}

		return false;
	}

	/**
	 * download_init( ):  The method sets some class variables
	 *                    The method checks, if download is allowed. User have to allow the download
	 *                    in the flexform / TypoScript
	 *                    The method explodes the given URL.
	 *                    Example:
	 *                    * URL is: single.301.tx_org_doc.9.documents.0
	 *                    * $this->view   = single
	 *                    * $this->mode   = 301
	 *                    * $this->table  = tx_org_doc
	 *                    * $this->uid    = 9
	 *                    * $this->field  = documents
	 *                    * $this->key    = 0
	 *
	 * @return	string		Prompt, in case of any downloading isn't allowed
	 * @version 3.9.3
	 * @since 3.9.3
	 */
	private function download_init()
	{

		$this->init_typo3version();
//////////////////////////////////////////////////////////////////////////
//
		// RETURN: Downloading isn\'t allowed

		$cObj_name = $this->pObj->conf[ 'flexform.' ][ 'sDEF.' ][ 'downloads.' ][ 'enabled' ];
		$cObj_conf = $this->pObj->conf[ 'flexform.' ][ 'sDEF.' ][ 'downloads.' ][ 'enabled.' ];
		$this->bool_downloadsAllowed = $this->pObj->cObj->cObjGetSingle( $cObj_name, $cObj_conf );

		if( !$this->bool_downloadsAllowed )
		{
			$prompt_01 = 'This current TYPO3 Browser plugin hasn\'t any permission to handle the download page object.';
			$prompt_02 = 'Please enable the property in the flexform tab [General] or in your TypoScript. See flexform.sDEF.downloads.enabled.';
			if( $this->pObj->b_drs_download )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 0 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/DOWNLOAD] ' . $prompt_02, $this->pObj->extKey, 1 );
			}
			return $prompt_01 . ' ' . $prompt_02;
		}
// RETURN: Downloading isn\'t allowed
//////////////////////////////////////////////////////////////////////////
//
		// Set all needed informations global
// piVars['file']: i.e. single.301.tx_org_doc.9.documents.0
		$arr_file = explode( '.', $this->pObj->piVars[ 'file' ] );
//var_dump( __METHOD__, __LINE__, $this->pObj->piVars );
//exit;
// view: list || single
		$this->view = $arr_file[ 0 ];
// mode (index) of the current view
		$this->mode = $arr_file[ 1 ];
// table: label of the current table
		$this->table = $arr_file[ 2 ];
// uid: uid of the current record
		$this->uid = ( int ) $arr_file[ 3 ];
// field: label of the field with the files
		$this->field = $arr_file[ 4 ];
// key: index of the file
		$this->key = ( int ) $arr_file[ 5 ];
// Set all needed informations global

		return;
	}

	/**
	 * getFile(): 
	 *
	 * @return	string
	 * @version 8.0.0
	 * @since 3.9.3
	 */
	private function getFile()
	{
//////////////////////////////////////////////////////////////////////////
//
		// Build and execute the query for getting files
// Build the query
		$uid = $this->pObj->objLocalise->get_localisedUid( $this->table, $this->uid );
		$select_fields = $this->field;
		$from_table = $this->table;
		$where_clause = 'uid = ' . $uid;
		$enablefields = $this->pObj->cObj->enableFields( $this->table );
		if( $enablefields )
		{
			$where_clause = $where_clause . $enablefields;
		}
		$query = $GLOBALS[ 'TYPO3_DB' ]->SELECTquery( $select_fields, $from_table, $where_clause, $groupBy = '', $orderBy = '', $limit = '' );

		if( $this->pObj->b_drs_sql || $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SQL+DOWNLOAD] ' . $query, $this->pObj->extKey, 0 );
		}
// Build the query
// Execute the query
		$res = $GLOBALS[ 'TYPO3_DB' ]->exec_SELECTquery( $select_fields, $from_table, $where_clause, $groupBy = '', $orderBy = '', $limit = '' );
// Execute the query
// Build and execute the query for getting files
//////////////////////////////////////////////////////////////////////////
//
		// Evaluate the query
// LOOP count rows of the SQL result
		$rows = array();
		$int_rows_counter = 0;
		// #i0366, 170919, dwildt, 1-/+
		//while( $row = $GLOBALS[ 'TYPO3_DB' ]->sql_fetch_assoc( $res ) )
		while( $row = $this->pObj->objSqlFun->sql_fetch_assoc( $res ) )
		{
			$rows[ $int_rows_counter ] = $row;
			$int_rows_counter++;
		}
// LOOP count rows of the SQL result
// Free the SQL result
		$GLOBALS[ 'TYPO3_DB' ]->sql_free_result( $res );

// RETURN: There are 0 ore more than one rows
		if( $int_rows_counter != 1 )
		{
			$prompt_01 = 'query: ' . $query;
			$prompt_02 = 'RETURN: Result are #' . $int_rows_counter . ' rows.';
			if( $this->pObj->b_drs_error )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_02, $this->pObj->extKey, 0 );
			}
			return $prompt_01 . '<br />' . $prompt_02;
		}
// RETURN: There are 0 ore more than one rows
// Evaluate the query
//////////////////////////////////////////////////////////////////////////
//
		// Get the file

		$files = explode( ',', $rows[ 0 ][ $this->field ] );
		$file = $files[ $this->key ];

		if( empty( $file ) )
		{
			$this->promptErrorAndDieNoFileKey( __METHOD__, __LINE__ );
		}

		return $file;
	}

	/**
	 * getPath( ) : Returns the absolute path
	 * 
	 * @return	string
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function getPath()
	{
		$uploadFolder = $this->getUploadfolder();


		$path = \TYPO3\CMS\Core\Utility\GeneralUtility::getFileAbsFileName( $uploadFolder );
		$path = rtrim( $path, '/' ) . '/';

		return $path;
	}

	/**
	 * getPathFileDAM(): 
	 *
	 * @return	string		pathToFile
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function getPathFileDAM()
	{
		$path = $this->getPath();
		$file = $this->getFile();

		$pathToFile = $path . $file;
		$this->promptErrorAndDieNoFileExist( $pathToFile, __METHOD__, __LINE__ );

		return $pathToFile;
	}

	/**
	 * getPathFileDefault(): 
	 *
	 * @return	string		pathToFile
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function getPathFileDefault()
	{
		$path = $this->getPath();
		$file = $this->getFile();

		$pathToFile = $path . $file;
		$this->promptErrorAndDieNoFileExist( $pathToFile, __METHOD__, __LINE__ );

		return $pathToFile;
	}

	/**
	 * getPathFileFAL(): 
	 *
	 * @return	string		$csvPaths
	 * @internal #t0933, #i0278
	 * @version 8.0.12
	 * @since 8.0.0
	 */
	private function getPathFileFAL()
	{

		$FAL = GeneralUtility::makeInstance( 'Netzmacher\\Browser\\Utility\\FAL' );

		$uid = $this->pObj->objLocalise->get_localisedUid( $this->table, $this->uid );
		$csvPaths = $FAL->getCsvPaths( $this->table, $this->field, $uid );
//		var_dump( __METHOD__, __LINE__, $table, $field, $this->row, $this->arrCategories, $csvPaths );
//		die();
		return $csvPaths;

//		$uid = $this->pObj->objLocalise->get_localisedUid( $this->table, $this->uid );
//
//    $fileRepository = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance('TYPO3\\CMS\\Core\\Resource\\FileRepository');
//    $fileObjects = $fileRepository->findByRelation($this->table, $this->field, $uid);
//
//    // get Imageobject information
//    $aPathToFile = array();
//    foreach ($fileObjects as $value) {
//      $aPathToFile[] = $value->getPublicUrl();
//    }
//    $csvPathToFile = implode(',', $aPathToFile);
//
////    var_dump(__METHOD__, __LINE__, $csvPathToFile);
////    die();
//
//    return $csvPathToFile;
	}

	/**
	 * getUploadfolder( ) : Get the upload folder.
	 *                                     Method takes the path to the upload folder from TCA.
	 *                                     In case of DAM (tx_dam) ist takes it from tx_dam.file_path
	 * @return	string     $uploadFolder  : The path to the upload folder
	 * @version 8.0.0
	 * @since 3.9.14
	 */
	private function getUploadfolder()
	{
		switch( TRUE )
		{
			case($this->caseDAM()):
				$uploadFolder = $this->getUploadfolderDAM();
				break;
			case($this->caseFAL()):
				$this->promptErrorAndDieUploadfolderFAL( __METHOD__, __LINE__ );
				break;
			default:
				$uploadFolder = $GLOBALS[ 'TCA' ][ $this->table ][ 'columns' ][ $this->field ][ 'config' ][ 'uploadfolder' ];
				break;
		}

// #t0784, 160725, dwildt, DEV!!!
		$uploadFolder = 'uploads/tx_org/';
		if( !empty( $uploadFolder ) )
		{
			return $uploadFolder;
		}

		$this->promptErrorAndDieUploadfolder( __METHOD__, __LINE__ );
	}

	/**
	 * getUploadfolderDAM( ) : 
	 * 
	 * @return	string     $uploadFolder  : The path to the upload folder
	 * @version 8.0.0
	 * @since 3.9.14
	 */
	private function getUploadfolderDAM()
	{
//////////////////////////////////////////////////////////////////////////
//
		// Build and execute the query for getting files
// Build the query
		$uid = $this->pObj->objLocalise->get_localisedUid( $this->table, $this->uid );
		$select_fields = 'file_path';
		$from_table = $this->table;
		$where_clause = 'uid = ' . $uid;
		$enablefields = $this->pObj->cObj->enableFields( $this->table );
		if( $enablefields )
		{
			$where_clause = $where_clause . $enablefields;
		}
		$query = $GLOBALS[ 'TYPO3_DB' ]->SELECTquery( $select_fields, $from_table, $where_clause, $groupBy = '', $orderBy = '', $limit = '' );
		if( $this->pObj->b_drs_sql || $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SQL+DOWNLOAD] ' . $query, $this->pObj->extKey, 0 );
		}
// Build the query
// Execute the query
		$res = $GLOBALS[ 'TYPO3_DB' ]->exec_SELECTquery( $select_fields, $from_table, $where_clause, $groupBy = '', $orderBy = '', $limit = '' );
// Execute the query
// Build and execute the query for getting files
//////////////////////////////////////////////////////////////////////////
//
		// Evaluate the query
// LOOP count rows of the SQL result
		$rows = array();
		$int_rows_counter = 0;
		// #i0366, 170919, dwildt, 1-/+
		//while( $row = $GLOBALS[ 'TYPO3_DB' ]->sql_fetch_assoc( $res ) )
		while( $row = $this->pObj->objSqlFun->sql_fetch_assoc( $res ) )
		{
			$rows[ $int_rows_counter ] = $row;
			$int_rows_counter++;
		}
// LOOP count rows of the SQL result
// Free the SQL result
		$GLOBALS[ 'TYPO3_DB' ]->sql_free_result( $res );

// RETURN: There are 0 ore more than one rows
		if( $int_rows_counter != 1 )
		{
			$prompt_01 = 'query: ' . $query;
			$prompt_02 = 'RETURN: Result are #' . $int_rows_counter . ' rows.';
			if( $this->pObj->b_drs_error )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_02, $this->pObj->extKey, 0 );
			}
			return $prompt_01 . '<br />' . $prompt_02;
		}
// RETURN: There are 0 ore more than one rows
// Evaluate the query
		$uploadFolder = $rows[ 0 ][ 'file_path' ];
//var_dump( __METHOD__, __LINE__, $uploadFolder );

		return $uploadFolder;
	}

	/**
	 * init_typo3version( ): Get the current TYPO3 version
	 *
	 * @return    integer
	 * @internal  #61520
	 * @version 6.0.0
	 * @since   4.5.11
	 */
	private function init_typo3version()
	{
		$this->typo3Version = $this->pObj->getTypo3Version();
	}

	/**
	 * promptErrorAndDieCase()
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function promptErrorAndDieCase( $case, $method, $line )
	{
		$prompt_01 = 'Sorry, the case \'' . $case . '\' is undefined.';
		$prompt_02 = $method . ' (#' . $line . ')';
		if( $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
		}

		echo ''
		. '<div style="border:1em solid red;text-align:center;padding:1em;">'
		. '<h1>TYPO3 Browser</h1>'
		. $prompt_01 . '<br />' . PHP_EOL
		. $prompt_02
		. '</div>';
		die();
	}

	/**
	 * promptErrorAndDieNoFileExist()
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function promptErrorAndDieNoFileExist( $file, $method, $line )
	{
		if( file_exists( $file ) )
		{
			return;
		}

		$prompt_01 = 'The file \'' . $file . '\' does not exist.';
		$prompt_02 = $method . ' (#' . $line . ')';
		if( $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
		}

		echo ''
		. '<div style="border:1em solid red;text-align:center;padding:1em;">'
		. '<h1>TYPO3 Browser</h1>'
		. $prompt_01 . '<br />' . PHP_EOL
		. $prompt_02
		. '</div>';
		die();
	}

	/**
	 * promptErrorAndDieNoFileExist()
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function promptErrorAndDieNoFileKey( $method, $line )
	{
		$prompt_01 = 'There isn\'t any file. Key is \'' . $this->key . '\'.';
		$prompt_02 = 'This error can happen, if the current download record contains a thumbnail but no document.';
		$prompt_03 = $method . ' (#' . $line . ')';
		if( $this->pObj->b_drs_error )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
		}

		echo ''
		. '<div style="border:1em solid red;text-align:center;padding:1em;">'
		. '<h1>TYPO3 Browser</h1>'
		. $prompt_01 . '<br />' . PHP_EOL
		. $prompt_02 . '<br />' . PHP_EOL
		. $prompt_03
		. '</div>';
		die();
	}

	/**
	 * promptErrorAndDieUploadfolder()
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function promptErrorAndDieUploadfolder( $method, $line )
	{
		$prompt_01 = 'Any upload folder is configured.';
		$prompt_02 = 'Please take care of a proper TCA configuration at: '
						. '$GLOBALS[ "TCA" ][ ' . $this->table . ' ][ "columns" ][ ' . $this->field . ' ][ "config" ][ "uploadfolder" ]'
		;
		$prompt_03 = 'Error in case of DAM: the field file_path is empty.';
		$prompt_04 = $method . ' (#' . $line . ')';
		if( $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/DOWNLOAD] ' . $prompt_02, $this->pObj->extKey, 1 );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/DOWNLOAD] ' . $prompt_02, $this->pObj->extKey, 1 );
		}

		echo ''
		. '<div style="border:1em solid red;text-align:center;padding:1em;">'
		. '<h1>TYPO3 Browser</h1>'
		. $prompt_01 . '<br />' . PHP_EOL
		. $prompt_02 . '<br />' . PHP_EOL
		. $prompt_03 . '<br />' . PHP_EOL
		. $prompt_04
		. '</div>';
		die();
	}

	/**
	 * promptErrorAndDieUploadfolderFAL()
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 8.0.0
	 * @since 8.0.0
	 */
	private function promptErrorAndDieUploadfolderFAL( $method, $line )
	{
		$prompt_01 = 'Case is FAL. But this isn\'t porper at this method.';
		$prompt_02 = $method . ' (#' . $line . ')';
		if( $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/DOWNLOAD] ' . $prompt_01, $this->pObj->extKey, 3 );
		}

		echo ''
		. '<div style="border:1em solid red;text-align:center;padding:1em;">'
		. '<h1>TYPO3 Browser</h1>'
		. $prompt_01 . '<br />' . PHP_EOL
		. $prompt_02
		. '</div>';
		die();
	}

	/*	 * *********************************************
	 *
	 * Sending
	 *
	 * ******************************************** */

	/**
	 * sendFileAndExit(): The method sends the file and exit in case of success
	 *                    The method checks:
	 *                    * upload folder is proper:  if there is a configuration in the TCA
	 *                    * sql result is proper:     if result is exactly one row
	 *                    * file is proper:           if the file does exist
	 *
	 * @return	string		Prompt, in case of a failure
	 * @version 3.9.14
	 * @since 3.9.3
	 */
	private function sendFileAndExit()
	{
		switch( TRUE )
		{
			case($this->caseDAM()):
				$pathToFile = $this->getPathFileDAM();
				break;
			case($this->caseFAL()):
				$pathToFile = $this->getPathFileFAL();
				break;
			case($this->caseDefault()):
				$pathToFile = $this->getPathFileDefault();
				break;
			default:
				$this->promptErrorAndDieCase( 'default', __METHOD__, __LINE__ );
				return;
		}

// RETURN: file doesn't exist
// Get the file
//////////////////////////////////////////////////////////////////////////
//
		// filefunc object
// #61520, 140911, dwildt, 2-
//      // Require fileFunc class
//    require_once( PATH_t3lib . 'class.t3lib_basicfilefunc.php' );
// #61520, 140911, dwildt, 4+
		if( $this->typo3Version < 6002000 )
		{
			require_once( PATH_t3lib . 'class.t3lib_basicfilefunc.php' );
		}
//// Initialize new fileFunc object
//    $this->fileFunc = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( 't3lib_basicFileFunctions' );
//// Get fileinfo
//    $fileInfo = $this->fileFunc->getTotalFileInfo( $pathToFile );

		$fileInfo = \TYPO3\CMS\Core\Utility\GeneralUtility::split_fileref( $pathToFile );
		$fileSize = filesize( $pathToFile );
		$file = $fileInfo[ 'filebody' ];

// filefunc object
//////////////////////////////////////////////////////////////////////////
//
		// Set the header

		$arr_header = null;

// Fileextension corresponds with a defined mimetype
		$str_fileext = $fileInfo[ 'fileext' ];
		if( isset( $this->pObj->conf[ 'download.' ][ 'mimetypes.' ][ 'fileext.' ][ $str_fileext ] ) )
		{
			$str_application = $this->pObj->conf[ 'download.' ][ 'mimetypes.' ][ 'fileext.' ][ $str_fileext ];
			$arr_header[ 'type' ] = 'Content-type: ' . $str_application;
		}
// Fileextension corresponds with a defined mimetype

		$arr_header[ 'description' ] = 'Content-Description: TYPO3 Browser Download Modul';
		$arr_header[ 'disposition' ] = 'Content-Disposition: attachment; filename="' . $file . '.' . $str_fileext . '"';
		$arr_header[ 'length' ] = 'Content-Length: ' . $fileSize;


// DRS - Development Reporting System
		if( $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] file name: ' . $file, $this->pObj->extKey, 0 );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] file path: ' . $fileInfo[ 'path' ], $this->pObj->extKey, 0 );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] file size: ' . $fileSize, $this->pObj->extKey, 0 );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] file extension: ' . $str_fileext, $this->pObj->extKey, 0 );
			$str_header = implode( ' // ', $arr_header );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/DOWNLOAD] header: ' . $str_header, $this->pObj->extKey, 0 );
		}
// DRS - Development Reporting System
// Loop header
		foreach( $arr_header as $str_header )
		{
			header( $str_header );
		}
// Loop header
// Set the header
//    $pos = strpos($this->pObj->str_developer_csvIp, \TYPO3\CMS\Core\Utility\GeneralUtility :: getIndpEnv('REMOTE_ADDR'));
//    if ( ! ( $pos === false ) )
//    {
//      var_dump(__METHOD__. ' (' . __LINE__ . '): ' , $fileInfo, $fileInfo['fileext'], $this->pObj->conf['download.']['mimetypes.']['fileext.'][$fileInfo['fileext']], $arr_header );
//      return;
//    }
//////////////////////////////////////////////////////////////////////////
//
		// Send the header and the file
// Read the file and write it to the output buffer.

		@readfile( $pathToFile ) || die( __METHOD__ . ' (' . __LINE__ . '): ' . readfile( $pathToFile ) );
		exit;
// Send the header and the file
	}

	/*	 * *********************************************
	 *
	 * typeNum
	 *
	 * ******************************************** */

	/**
	 * set_typeNum(): Set the class variables $int_typeNum and $str_typeNum.
	 *                The class variables are needed by other classes while runtime.
	 *
	 * @return	void
	 * @version 3.9.3
	 * @since 3.9.3
	 */
	public function set_typeNum()
	{
// Get the typeNum from the current URL parameters
		$typeNum = ( int ) \TYPO3\CMS\Core\Utility\GeneralUtility::_GP( 'type' );

// RETURN typeNum is 0 or empty
		if( empty( $typeNum ) )
		{
			if( $this->pObj->b_drs_download )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devLog( '[INFO/DOWNLOAD] typeNum is 0 or empty.', $this->pObj->extKey, 0 );
			}
			return;
		}
// RETURN typeNum is 0 or empty
// Check the proper typeNum
		$conf = $this->pObj->conf;
		switch( true )
		{
			case( $typeNum == $conf[ 'download.' ][ 'page.' ][ 'typeNum' ] ) :
				// Given typeNum is the internal typeNum for download
				$this->int_typeNum = $typeNum;
				$this->str_typeNum = 'download';
				break;
			default :
				// Given typeNum isn't the internal typeNum for download
				$this->str_typeNum = 'undefined';
		}
		unset( $conf );
// Check the proper typeNum
// DRS - Development Reporting System
		if( $this->pObj->b_drs_download )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devLog( '[INFO/DOWNLOAD] typeNum is \'' . $typeNum . '\'. Name is \'' . $this->str_typeNum . '\'.', $this->pObj->extKey, 0 );
		}
// DRS - Development Reporting System
	}

	/*	 * *********************************************
	 *
	 * Statistics
	 *
	 * ******************************************** */

	/**
	 * download_statistics(): Inits the statistic module
	 *
	 * @param	string		$operator: operator has to be + or -
	 * @return	void
	 * @version 3.9.3
	 * @since 3.9.3
	 */
	private function statistics( $operator )
	{
//////////////////////////////////////////////////////////////////////////
//
		// Set status of the statistics module and init it

		$this->pObj->objStat->statisticsIsEnabled();
// Set status of the statistics module and init it
//////////////////////////////////////////////////////////////////////////
//
		// RETURN: statistics module is disabled

		if( !$this->pObj->objStat->bool_statistics_enabled )
		{
			if( $this->pObj->b_drs_statistics )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STATISTICS] single view won\'t counted for statistics.', $this->pObj->extKey, 0 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/STATISTICS] Enable flexform.sDEF.statistics.enabled.', $this->pObj->extKey, 1 );
			}
			return;
		}
// RETURN: statistics module is disabled



		$this->statistics_download( $operator );
		$this->statistics_downloadByVisit( $operator );
	}

	/**
	 * statistics_download(): Count the hit
	 *
	 * @param	string		$operator: operator has to be + or -
	 * @return	void
	 * @version 3.9.3
	 * @since 3.9.3
	 */
	private function statistics_download( $operator )
	{
		$field = $this->pObj->objStat->fieldDownloads;
// Count the hit
		$this->pObj->objStat->sql_update_statistics( $this->table, $field, $this->uid, $operator );
	}

	/**
	 * statistics_downloadByVisit(): Count the hit by visit
	 *
	 * @param	string		$operator: operator has to be + or -
	 * @return	void
	 * @version 3.9.3
	 * @since 3.9.3
	 */
	private function statistics_downloadByVisit( $operator )
	{
		$field = $this->pObj->objStat->fieldDownloadsByVisits;
// RETURN: no new visit
//$bool_newVisit = $this->pObj->objSession->statisticsNewDownload( $this->table, $field, $this->uid );
		$bool_newVisit = $this->pObj->objSession->statisticsNewVisit( $this->table, $field, $this->uid );
		if( !$bool_newVisit )
		{
// DRS - Development Reporting System
			if( $this->pObj->b_drs_statistics )
			{
				$prompt = 'No new visit, no counting.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STATISTICS] ' . $prompt, $this->pObj->extKey, 0 );
			}
// DRS - Development Reporting System
			return;
		}
// RETURN: no new visit
// Count the hit
		$field = $this->pObj->objStat->fieldDownloadsByVisits;
		$this->pObj->objStat->sql_update_statistics( $this->table, $field, $this->uid, $operator );

		return;
	}

}

if( defined( 'TYPO3_MODE' ) && $TYPO3_CONF_VARS[ TYPO3_MODE ][ 'XCLASS' ][ 'ext/browser/Classes/Plugins/Pi1/class.tx_browser_pi1_download.php' ] )
{
	include_once ($TYPO3_CONF_VARS[ TYPO3_MODE ][ 'XCLASS' ][ 'ext/browser/Classes/Plugins/Pi1/class.tx_browser_pi1_download.php' ]);
}