<?php

use \TYPO3\CMS\Core\Utility\GeneralUtility;
use Netzmacher\Browser\Utility\UserFunc\PhpHelper;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2010-2019 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * The class tx_browser_pi1_marker bundles marker methods for the extension browser
 *
 * @author    Dirk Wildt http://wildt.at.die-netzmacher.de
 *
 * @package    TYPO3
 * @subpackage   browser
 * @version   8.0.64
 * @since     3.4.4
 */
class tx_browser_pi1_marker
{

	// [Array] The current TypoScript configuration array
	var $conf = false;
	// [Array] Temporarily array for storing piVars
	var $tmp_piVars = false;
	// [Array] Array with all keys of the TYPO3 array $GLOBALS
	var $arr_t3global_keys = false;

	/**
	 * Constructor. The method initiate the parent object
	 *
	 * @param    object        The parent object
	 * @access public
	 * @return    void
	 */
	public function __construct( $parentObj )
	{
		$this->pObj = $parentObj;
		$this->conf = $this->pObj->conf;
	}

	/**
	 * _elements_loop( ):
	 *
	 * @param
	 * @return    string      $value_tsConf_after_loop
	 * @access private
	 * @version   8.0.48
	 * @since     4.4.0
	 */
	private function _elements_loop( $elements, $value_tsConf_after_loop, $key_tsConf )
	{
// DRS
		if( $this->pObj->boolFirstRow && $this->pObj->b_drs_marker )
		{
			$prompt = '$elements: ' . var_export( $elements, true );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ' . $prompt, $this->pObj->extKey, 0 );
		}
// DRS
// FOREACH  : elements
		foreach( ( array ) $elements as $key_tableField => $value_tableField )
		{
			// cHashMarker
			$key_marker = '###' . strtoupper( $key_tableField ) . '###';

			// #44316, 130105, dwildt, TODO: Uncomment the code below?
			// session marker
//        if(in_array('session.', $key_tsConf))
//        {
//            // 110124, dwildt, :todo: session
//          $elements = $this->session_marker($value_tsConf_after_loop, $elements);
//        }
			// session marker
			// CONTINUE : Value doesn't contain the current marker
			if( strpos( $value_tsConf_after_loop, $key_marker ) === false )
			{
				continue;
			}
			// CONTINUE : Value doesn't contain the current marker
			// CONTINUE : Marker hasn't any child value, replace marker with value
			if( !in_array( $key_tsConf, ( array ) $this->pObj->arr_children_to_devide ) )
			{
				// Color swords
				$value_tableField = $this->pObj->objZz->color_swords( $value_tableField );
				$value_tsConf_after_loop = str_replace( $key_marker, $value_tableField, $value_tsConf_after_loop );
				continue;
			}
			// CONTINUE : Marker hasn't any child value, replace marker with value
			// Replace marker with children
			$value_tsConf_after_loop = $this->_elements_loopReplaceChildren( $value_tsConf_after_loop, $value_tableField, $key_marker );
		}
// FOREACH  : elements

		return $value_tsConf_after_loop;
	}

	/**
	 * _elements_loopReplaceChildren( ):
	 *
	 * @param
	 * @return    string      $value_tsConf_after_loop
	 * @access private
	 * @version   8.0.48
	 * @since     4.4.0
	 */
	private function _elements_loopReplaceChildren( $value_tsConf_after_loop, $value_tableField, $key_marker )
	{
//    $this->pObj->dev_var_dump( 'X' );
// Get the sqlDevider configuration
		$str_devider = $this->pObj->objTyposcript->set_confSqlDevider();
		$str_sqlDeviderDisplay = $this->pObj->objTyposcript->str_sqlDeviderDisplay;

//      // #44316, 130105, dwildt, 7+
//      // DRS
//    if( $this->pObj->boolFirstRow && $this->pObj->b_drs_error )
//    {
//      $prompt = 'It seem\'s, that children aren\'t run proper!';
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/MARKER] ' . $prompt, $this->pObj->extKey, 3 );
//    }
//      // DRS
// Get children values
		$arr_valuesChildren = explode( $str_devider, $value_tableField );

// Multiple the values and replace the marker for every child
// EXAMPLE for value
//   Before marker replacement: &tx_trevent_pi1[uid]=###FE_USERS.UID###&###CHASH###
//   After  marker replacement: &tx_trevent_pi1[uid]=158&###CHASH###, ;|;&tx_trevent_pi1[uid]=155&###CHASH###
		$arr_value_after_loop = null;
		foreach( ( array ) $arr_valuesChildren as $valueChild )
		{
			$arr_value_after_loop[] = str_replace( $key_marker, $valueChild, $value_tsConf_after_loop );
		}
// 13008, 110302, dwildt
// 13807, 110313, dwildt
		$value_tsConf_after_loop = implode( $str_sqlDeviderDisplay, ( array ) $arr_value_after_loop );
// Multiple the values and replace the marker for every child

		return $value_tsConf_after_loop;
	}

	/**
	 * _extend_elements( ):
	 *
	 * @param     array       $elements : The current row of the SQL result
	 * @return    array       $elements : extended with piVars and siglePid
	 * @access  private
	 * @version   8.0.48
	 * @since     4.4.0
	 */
	private function _extend_elements( $elements )
	{

		$elements = $this->_extend_elementsWiPiVars( $elements );
		if( empty( $elements ) )
		{
			return;
		}

		$elements = $this->_extend_elementsWiSinglePid( $elements );

		return $elements;
	}

	/**
	 * _extend_elementsWiPiVars( ):
	 *
	 * @param     array       $elements : The current row of the SQL result
	 * @return    array       $elements : extended with piVars and siglePid
	 * @access  private
	 * @version   8.0.48
	 * @since     4.4.0
	 */
	private function _extend_elementsWiPiVars( $elements )
	{
		foreach( ( array ) $this->pObj->piVars as $key => $value )
		{
			// dwildt, 090620: If we have multiple selects, piVars can contain arrays
			// This array should be array of uids. We don't need any process for uids here.
			if( !is_array( $value ) )
			{
				// dwildt, 110320: Prevent to override database values in $elements by piVars
				if( !isset( $elements[ strtolower( $key ) ] ) )
				{
					$elements[ strtolower( $key ) ] = $value;
				}
			}
			if( $this->pObj->boolFirstRow && $this->pObj->b_drs_marker )
			{
				$prompt = 'The piVar [' . $key . '] is available.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ' . $prompt, $this->pObj->extKey, 0 );
				$prompt = 'If you use the marker ###' . strtoupper( $key ) . '###, it will become ' . $value;
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/MARKER] ' . $prompt, $this->pObj->extKey, 1 );
			}
		}

		return $elements;
	}

	/**
	 * _extend_elementsWiSinglePid( ):
	 *
	 * @param     array       $elements : The current row of the SQL result
	 * @return    array       $elements : extended with piVars and siglePid
	 * @access  private
	 * @version   8.0.48
	 * @since     4.4.0
	 */
	private function _extend_elementsWiSinglePid( $elements )
	{

		if( !isset( $this->pObj->singlePid ) )
		{
			return $elements;
		}

		$elements[ strtolower( 'singlePid' ) ] = $this->pObj->singlePid;
		return $elements;
	}

	/**
	 * extend_marker_wi_cObjData: Extend the given marker array with key/values from cObj->data,
	 *                            the data of the tt_content record of the browser plugin
	 *
	 * @param    array        $markerArray: Array with markers
	 * @return    array        $markerArray: Array with markers extended
	 * @version 5.0.0
	 * @since 3.7.0
	 */
	private function _extend_marker_wi_cObjData( $markerArray )
	{
		foreach( ( array ) $this->pObj->cObj->data as $key_cObjData => $value_cObjData )
		{
			if( !empty( $value_cObjData ) )
			{
				$markerArray[ '###TT_CONTENT.' . strtoupper( $key_cObjData ) . '###' ] = $value_cObjData;
				if( $this->pObj->boolFirstRow && $this->pObj->b_drs_marker )
				{
					// Make serial array line-break-able
					$value_cObjData = str_replace( ';', '; ', $value_cObjData );
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] The cObjData [' . $key_cObjData . '] is available.', $this->pObj->extKey, 0 );
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/MARKER] If you use the marker ###TT_CONTENT.' . strtoupper( $key_cObjData ) . '###, it will become ' . $value_cObjData, $this->pObj->extKey, 1 );
				}
			}
		}

		return $markerArray;
	}

	/**
	 * extend_marker_wi_pivars: Extend the given marker array with key/values from the piVars
	 *
	 * @param    array        $markerArray: Array with markers
	 * @return    array        $markerArray: Array with markers extended
	 * @version 5.0.0
	 * @since 2.0.0
	 */
	private function _extend_marker_wi_pivars( $markerArray )
	{

//var_dump( __METHOD__, __LINE__, $markerArray['###TX_ORG_HEADQUARTERS.TITLE###']);
//var_dump( __METHOD__, __LINE__, $this->pObj->piVars );

		foreach( ( array ) $this->pObj->piVars as $key_pivar => $value_pivar )
		{
			if( !empty( $value_pivar ) )
			{
				$markerArray[ '###' . strtoupper( $key_pivar ) . '###' ] = $value_pivar;
				if( $this->pObj->boolFirstRow && $this->pObj->b_drs_marker )
				{
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] The piVar [' . $key_pivar . '] is available.', $this->pObj->extKey, 0 );
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/MARKER] If you use the marker ###' . strtoupper( $key_pivar ) . '###, it will become ' . $value_pivar, $this->pObj->extKey, 1 );
				}
			}
		}

		return $markerArray;
	}

	/**
	 * _getConstantMarkers(): Generate the markerArray with self-defined markers out of the TypoScript. Return a markerArray, if there are values for replacement.
	 *
	 * @return	array		The markerArray. If there aren't any value, it returns FALSE.
	 * @version 8.0.64
	 * @internal #i0381.8.0.64. This method is taken from pi1/class.tx_browser_pi1_wrapper4x.php
	 * @since 1.0.0
	 * @internal :TODO: 171009: replacement of tablefield-marker below isn't needed
	 */
	private function _getConstantMarkers( $row = null )
	{
		$conf = $this->pObj->conf;
		$mode = $this->pObj->piVar_mode;
		$view = $this->pObj->view;

		$viewWiDot = $view . '.';
		$conf_view = $conf[ 'views.' ][ $viewWiDot ][ $mode . '.' ];

		// #i0050, 140630, dwildt, 4+
		if( $row === null )
		{
			$row = $this->pObj->elements;
		}

		// Get the TypoScript marker array. RETURN, if there isn't any array.
		$conf_marker = $conf_view[ 'marker.' ];
		if( !is_array( $conf_marker ) )
		{
			if( $this->pObj->b_drs_templating )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devLog( '[INFO/TEMPLATING] views.' . $viewWiDot . $mode . ' hasn\'t any marker array. We take the global one.', $this->pObj->extKey, 0 );
			}
			$conf_marker = $conf[ 'marker.' ];
			if( !is_array( $conf_marker ) )
			{
				if( $this->pObj->b_drs_templating )
				{
					\TYPO3\CMS\Core\Utility\GeneralUtility::devLog( '[INFO/TEMPLATING] The plugin hasn\'t any marker array. Self defined markers in HTML templates won\'t be replaced.', $this->pObj->extKey, 0 );
				}
				return false;
			}
		}

		$conf_marker = $this->pObj->objMarker->substitute_tablefield_marker( $conf_marker );
		// #t1649, 171008, dwildt, 1-
//		$conf_oneDim_marker = TYPO3\CMS\Backend\Utility\BackendUtility::implodeTSParams( $conf_marker );
		// #t1649, 171008, dwildt, 1+
		$PhpHelper = GeneralUtility::makeInstance( PhpHelper::class );

		// Loop through all elements (real values)
		foreach( ( array ) $row as $key_tableField => $value_tableField )
		{
			$arrSearch[] = '###' . strtoupper( $key_tableField ) . '###';
			$arrReplace[] = $value_tableField;

			// #t1649, 171008, dwildt, -
//			// Loop through one dimensional marker array
//			foreach( ( array ) $conf_oneDim_marker as $key => $value )
//			{
//				// Replace constant marker with real value
//				$value = str_replace( '###' . strtoupper( $key_tableField ) . '###', $value_tableField, $value );
//				// cHash marker
//				$pos = strpos( $value, '&###CHASH###' );
//				if( !($pos === false) )
//				{
//					$str_path = str_replace( '&###CHASH###', '', $value );
//					$arr_url = parse_url( $str_path );
//					$cHash_md5 = $this->pObj->objZz->get_cHash( $arr_url[ 'path' ] );
//					$value = str_replace( '&###CHASH###', '&cHash=' . $cHash_md5, $value );
//				}
//				// cHash marker
//				// session marker
//				$conf_oneDim_marker[ $key ] = $value;
//			}
		}


		// #t1649, 171008, dwildt, 3-
//		unset( $conf_marker );
//		// Rebild tsConf marker
//		$conf_marker = $this->pObj->objTyposcript->oneDim_to_tree( $conf_oneDim_marker );
		// #t1649, 171008, dwildt, 1+
		$PhpHelper->replaceInArrayTree( $arrSearch, $arrReplace, $conf_marker );
		//die( __METHOD__ . '#' . __LINE__ );
		// #12472, 110124, dwildt
		//////////////////////////////////////////////////////////
		//
      // Building the marker array for replacement
		// 130104, dwildt, 1-
//    foreach((array) $conf_marker as $key_marker => $arr_marker)
		// 130104, dwildt, 1+
		foreach( array_keys( ( array ) $conf_marker ) as $key_marker )
		{
			if( substr( $key_marker, -1, 1 ) == '.' )
			{
				// I.e. $key_marker is 'title.', but we like the marker name without any dot
				$str_marker = substr( $key_marker, 0, strlen( $key_marker ) - 1 );
				$coa_name = $conf_marker[ $str_marker ];
				if( empty( $coa_name ) )
				{
					$coa_name = 'TEXT';
				}
				$coa_conf = $conf_marker[ $str_marker . '.' ];
				$value = $this->pObj->cObj->cObjGetSingle( $coa_name, $coa_conf );
				$hashKeyMarker = '###' . strtoupper( $str_marker ) . '###';
				$markerArray[ $hashKeyMarker ] = $value;
			}
		}
		// Building the marker array for replacement
		//////////////////////////////////////////////////////////////////////
		//
      // AJAX
		// #9659, 101010 fsander
		if( $this->pObj->objFlexform->bool_ajax_enabled )
		{
			$markerArray[ '###BROWSER_ID###' ] = $this->pObj->cObj->data[ 'uid' ];
			if( $this->pObj->b_drs_template || $this->pObj->b_drs_javascript )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/TEMPLATING+JSS] AJAX: marker_array is extended with ###BROWSER_ID###: ' .
								$this->pObj->cObj->data[ 'uid' ], $this->pObj->extKey, 0 );
			}
		}
		// AJAX

		return $markerArray;
	}

	/**
	 * _replaceCHash():  
	 *
	 * @param    string
	 * @return    string
	 * @access private
	 * @version 8.0.48
	 * @since   8.0.48
	 */
	private function _replaceCHash( $value )
	{
		$pos = strpos( $value, '&###CHASH###' );
		if( ( $pos === false ) )
		{
			return $value;
		}

		$str_path = str_replace( '&###CHASH###', '', $value );
		$arr_url = parse_url( $str_path );
		$cHash_md5 = $this->pObj->objZz->get_cHash( $arr_url[ 'path' ] );
		$value = str_replace( '&###CHASH###', '&cHash=' . $cHash_md5, $value );
		return $value;
	}

	/**
	 * _replaceCHash():  
	 *
	 * @param    array
	 * @param    string
	 * @return    array
	 * @access private
	 * @version 8.0.48
	 * @since   8.0.48
	 */
	private function _resetFirstKey( $conf, $properKey )
	{
		$unproperKey = key( $conf );
		$firstElement[ $properKey ] = $conf[ $unproperKey ];
		unset( $conf[ $unproperKey ] );
		$conf = $firstElement + $conf;

		return $conf;
	}

	/**
	 * _substitute_tablefield_marker_loop():  
	 *
	 * @param    array        $arr_multi_dimensional: Multi-dimensional array like an TypoScript array
	 * @param    array        $elements: The current row of the SQL result
	 * @return    array        $arr_multi_dimensional: The current Multi-dimensional array with substituted markers
	 * @access private
	 * @version 8.0.48
	 * @since   3.6.0
	 */
	private function _substitute_tablefield_marker_loop( $confOneDimension, $key, $value, $elements )
	{
		$newValue = $value;

		$countHashMarker = substr_count( $value, '###' ); // I.e: 4

		if( $countHashMarker == 0 )
		{
			if( !$this->pObj->boolFirstRow OR ! $this->pObj->b_drs_marker )
			{
				return;
			}
			$this->_zzDRSWoHash( $value );
			return;
		}

		$newValue = $this->_elements_loop( $elements, $newValue, $key );
		$newValue = $this->_replaceCHash( $newValue );
		$newValue = $this->replace_left_over( $newValue );
		$confOneDimension[ $key ] = $newValue;

		$this->_zzDRSWiHash( $newValue, $value, $key );

		return $confOneDimension;
	}

	/**
	 * _zzDRSWiHash( )  
	 *
	 * @param    string
	 * @param    string
	 * @return    void  
	 * @access private
	 * @version 8.0.48
	 * @since   8.0.48
	 */
	private function _zzDRSWiHash( $newValue, $value, $key )
	{
		if( !$this->pObj->boolFirstRow OR ! $this->pObj->b_drs_marker )
		{
			return;
		}
		if( $newValue == $value )
		{
			return;
		}

		$value_html = htmlspecialchars( $newValue );
		if( strlen( $value_html ) > $this->pObj->i_drs_max_sql_result_len )
		{
			$value_html = substr( $value_html, 0, $this->pObj->i_drs_max_sql_result_len ) . ' ...';
		}

		if( empty( $value_html ) )
		{
			$prompt = '... [' . $key . ']: ' . $value . ' is EMPTY.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ' . $prompt, $this->pObj->extKey, 0 );
		}
		if( !empty( $value_html ) )
		{
			$prompt = '... [' . $key . ']: ' . $value . ' become: ' . $value_html;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ' . $prompt, $this->pObj->extKey, 0 );
		}
	}

	/**
	 * _zzDRSWoHash( )  
	 *
	 * @param    string
	 * @return    void  
	 * @access private
	 * @version 8.0.48
	 * @since   8.0.48
	 */
	private function _zzDRSWoHash( $value )
	{
		$value_html = htmlspecialchars( $value );
		if( strlen( $value_html ) > $this->pObj->i_drs_max_sql_result_len )
		{
			$value_html = substr( $value_html, 0, $this->pObj->i_drs_max_sql_result_len ) . ' ...';
		}
		$prompt = 'without marker: ' . $value;
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ' . $prompt, $this->pObj->extKey, 0 );
	}

	/**
	 * extend_marker(): 
	 *
	 * @return	array   $markerArray
	 * @access public
	 * @version 8.0.64
	 * @since 5.0.0
	 */
	public function extend_marker( $markerArray = array() )
	{
		$markerArray[ '###MODE###' ] = $this->pObj->piVar_mode;
		$markerArray[ '###VIEW###' ] = $this->pObj->view;
		$markerArray = $this->_extend_marker_wi_cObjData( $markerArray );
		$markerArray = $this->_extend_marker_wi_pivars( $markerArray );
		// #i0383.8.0.66, 171121, dwildt, 2-/5+		
//		// #i0381.8.0.64, 171103, dwildt, 1+/5-
//		$markerArray = $this->extend_marker_wiConstantMarkers( $markerArray );
		$constantMarkers = $this->pObj->objWrapper4x->constant_markers();
		foreach( ( array ) $constantMarkers as $key => $value )
		{
			$markerArray[ $key ] = $value;
		}
		return $markerArray;
	}
	/**
	 * extend_marker(): 
	 *
	 * @return	array   $markerArray
	 * @access public
	 * @version 8.0.48
	 * @since 5.0.0
	 */
	public function extend_marker_wiConstantMarkers( $markerArray )
	{
		$constantMarkers = $this->_getConstantMarkers();
		foreach( ( array ) $constantMarkers as $key => $value )
		{
			$markerArray[ $key ] = $value;
		}
		return $markerArray;
	}

	/**
	 * replace_left_over(): Replace all markers, which are left over
	 *                      Feature: #28657
	 *
	 * @param    array        $str_content: current content
	 * @return    string        $str_content: rendered content
	 * @access public
	 * @version 8.0.48
	 * @since 3.7.0
	 */
	public function replace_left_over( $str_content )
	{

// Get configuration from the flexform
// Bugfix     #29738, uherrmann, 110913
		$sheet = 'evaluate';
		$field = 'handle_marker';
		// #t1750, 171101, dwildt, 1-/+
		//$arr_piFlexform = $this->pObj->cObj->data[ 'pi_flexform' ];
		$arr_piFlexform = GeneralUtility::xml2array($this->pObj->cObj->data[ 'pi_flexform' ]);

		$handle_marker = $this->pObj->pi_getFFvalue( $arr_piFlexform, $field, $sheet, 'lDEF', 'vDEF' );

// Switch configuration
		switch( $handle_marker )
		{
			case('remove_empty_markers'):
			case(''):
			case(null):
				// DRS - Development Reporting System
				if( $this->pObj->b_drs_marker )
				{
					// 130105, dwildt, 1+
					$arr_markers = array();
					preg_match_all
									(
									'|###.*?###|i', $str_content, $arr_markers, PREG_PATTERN_ORDER
					);
					$str_markers = implode( ', ', $arr_markers[ 0 ] );
					\TYPO3\CMS\Core\Utility\GeneralUtility::devLog( '[INFO/TEMPLATE] Markers will be removed: ' . $str_markers, $this->pObj->extKey, 0 );
				}
				// DRS - Development Reporting System
				// Replace the left over markers
				$str_content = preg_replace( '|###.*?###|i', '', $str_content );
				break;
			case('none'):
				// Do nothing;
				break;
			default:
				$prompt = '
                  <div style="background:white; font-weight:bold;border:.4em solid orange;">
                    <h1>
                      WARNING
                    </h1>
                    <p>
                      Flexform field has an invalid value. The value isn\'t defined.<br />
                      sheet: ' . $sheet . '<br />
                      field: ' . $field . '<br />
                      value: ' . $handle_marker . '<br />
                      ' . __METHOD__ . ' (' . __LINE__ . ')
                    </p>
                    <p>
                      Please save the plugin/flexform of the browser. The bug will be fixed probably.
                    </p>
                  </div>';
				echo $prompt;
		}
// Switch configuration

		return $str_content;
	}

	/**
	 * substitute_marker():  Replace markers:
	 *                       Replace all markers in the given multidimensional array like the TypoScript configuration
	 *                       with the values from the given marker array
	 *                                  * The method extends the SQL result with all piVar values. ###CHASH### has a process.
	 *                                  * This method should supersede the deprecated method _substitute_marker_recursive ()
	 *
	 * @param    array        $arr_multi_dimensional: Multi-dimensional array like an TypoScript array
	 * @param    array        $elements: The current row of the SQL result
	 * @return    array        $arr_multi_dimensional: The current Multi-dimensional array with substituted markers
	 * @access public
	 * @version 8.0.48
	 * @since   3.6.0
	 */
	public function substitute_marker( $arr_multi_dimensional, $marker )
	{
//////////////////////////////////////////////////////////////////////////
//
		// RETURN there isn't any element

		if( empty( $marker ) )
		{
			return $arr_multi_dimensional;
		}
// RETURN there isn't any element
//////////////////////////////////////////////////////////////////////////
//
		// Add to the $marker piVars and singlePid
// Add to the $marker piVars
		foreach( $this->pObj->piVars as $key_pivar => $value_pivar )
		{
			// dwildt, 090620: If we have multiple selects, piVars can contain arrays
			// This array should be array of uids. We don't need any process for uids here.
			if( !is_array( $value_pivar ) )
			{
				// dwildt, 110320: Prevent to override database values in $marker by piVars
				if( !isset( $marker[ '###' . strtoupper( $key_pivar ) . '###' ] ) )
				{
					$marker[ '###' . strtoupper( $key_pivar ) . '###' ] = $value_pivar;
				}
			}
			if( $this->pObj->boolFirstRow && $this->pObj->b_drs_marker )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] The piVar [' . $key_pivar . '] is available.', $this->pObj->extKey, 0 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/MARKER] If you use the marker ###' . strtoupper( $key_pivar ) . '###, it will become ' . $value_pivar, $this->pObj->extKey, 1 );
			}
		}
// Add to the $marker piVars
// Add to the $marker the singlePid
		if( isset( $this->pObj->singlePid ) )
		{
			$marker[ '###' . strtoupper( 'singlePid' ) . '###' ] = $this->pObj->singlePid;
		}
// Add to the $marker the singlePid
// Add to the $marker piVars and singlePid
//////////////////////////////////////////////////////////////////////////
//
		// Loop through one dimensional array

		$arr_one_dimensional = Netzmacher\Refresh\Compatibility\Backend\Utility\BackendUtility::implodeTSParams( $arr_multi_dimensional );
		foreach( ( array ) $arr_one_dimensional as $key_tsConf => $value_tsConf )
		{
			$value_tsConf_after_loop = $value_tsConf;

			// CONTINUE: array element dosn't contain any marker
			$int_countMarker = substr_count( $value_tsConf, '###' ); // I.e: 4
			if( $int_countMarker == 0 )
			{
				if( $this->pObj->b_drs_marker )
				{
					$value_tsConf_html = htmlspecialchars( $value_tsConf );
					if( strlen( $value_tsConf_html ) > $this->pObj->i_drs_max_sql_result_len )
					{
						$value_tsConf_html = substr( $value_tsConf_html, 0, $this->pObj->i_drs_max_sql_result_len ) . ' ...';
					}
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ... ' . $value_tsConf . ' hasn\'t any marker.', $this->pObj->extKey, 0 );
				}
				continue;
			}
			// CONTINUE: array element dosn't contain any marker
			// Replace markers
			foreach( ( array ) $marker as $key_marker => $value_marker )
			{
				$value_tsConf_after_loop = str_replace( $key_marker, $value_marker, $value_tsConf_after_loop );
			}
			// Replace markers
			// Replace cHash marker
			$pos = strpos( $value_tsConf_after_loop, '&###CHASH###' );
			if( !($pos === false) )
			{
				$str_path = str_replace( '&###CHASH###', '', $value_tsConf_after_loop );
				$arr_url = parse_url( $str_path );
				$cHash_md5 = $this->pObj->objZz->get_cHash( $arr_url[ 'path' ] );
				$value_tsConf_after_loop = str_replace( '&###CHASH###', '&cHash=' . $cHash_md5, $value_tsConf_after_loop );
			}
			// Replace cHash marker
			// Clear markers, which aren't replaced
			$value_tsConf_after_loop = preg_replace( '|###.*?###|i', '', $value_tsConf_after_loop );

			// DRS - Development Reporting System
			if( $value_tsConf_after_loop != $value_tsConf )
			{
				if( $this->pObj->b_drs_marker )
				{
					$value_tsConf_html = htmlspecialchars( $value_tsConf_after_loop );
					if( strlen( $value_tsConf_html ) > $this->pObj->i_drs_max_sql_result_len )
					{
						$value_tsConf_html = substr( $value_tsConf_html, 0, $this->pObj->i_drs_max_sql_result_len ) . ' ...';
					}
					if( empty( $value_tsConf_html ) )
					{
						\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ... [' . $key_marker . ']: ' . $value_tsConf . ' is EMPTY.', $this->pObj->extKey, 0 );
					}
					if( !empty( $value_tsConf_html ) )
					{
						\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/MARKER] ... [' . $key_marker . ']: ' . $value_tsConf . ' become:<br /><br />' . $value_tsConf_html, $this->pObj->extKey, 0 );
					}
				}
			}
			// DRS - Development Reporting System

			$arr_one_dimensional[ $key_tsConf ] = $value_tsConf_after_loop;
		}
// Loop through one dimensional array
//////////////////////////////////////////////////////////////////////////
//
		// Rebuild multidimensional array

		unset( $arr_multi_dimensional );
		$arr_multi_dimensional = $this->pObj->objTyposcript->oneDim_to_tree( $arr_one_dimensional );
// Rebuild multidimensional array
// RETURN multidimensional array
		return $arr_multi_dimensional;
	}

	/**
	 * substitute_tablefield_marker():  Replace database markers:
	 *                                  Replace all markers in the given multidimensional array like the TypoScript
	 *                                  configuration with the real values from the SQL result (with table.field values)
	 *                                  * The method extends the SQL result with all piVar values. ###CHASH### has a process.
	 *                                  * This method should supersede the deprecated method _substitute_marker_recursive ()
	 *
	 * @param    array        $arr_multi_dimensional: Multi-dimensional array like an TypoScript array
	 * @param    array        $elements: The current row of the SQL result
	 * @return    array        $arr_multi_dimensional: The current Multi-dimensional array with substituted markers
	 * @access public
	 * @version 5.0.0
	 * @since   3.6.0
	 */
	public function substitute_tablefield_marker( $conf, $elements = null )
	{
		if( $this->conf[ 'performance.' ][ 'substitute_tablefield_marker' ] )
		{
			return $conf;
		}

		// Extend elements with piVars and singlePid
		$elements = $this->_extend_elements( $elements );
		if( empty( $elements ) )
		{
			return $conf;
		}

		$properKey = key( $conf );
		$confOneDimension = Netzmacher\Refresh\Compatibility\Backend\Utility\BackendUtility::implodeTSParams( $conf );

		foreach( ( array ) $confOneDimension as $key_tsConf => $value_tsConf )
		{
			$confOneDimension = $this->_substitute_tablefield_marker_loop( $confOneDimension, $key_tsConf, $value_tsConf, $elements );
		}

		unset( $conf );

		$conf = $this->pObj->objTyposcript->oneDim_to_tree( $confOneDimension );
		$conf = $this->_resetFirstKey( $conf, $properKey );

		return $conf;
	}

}

if( defined( 'TYPO3_MODE' ) && $TYPO3_CONF_VARS[ TYPO3_MODE ][ 'XCLASS' ][ 'ext/browser/Classes/Plugins/Pi1/class.tx_browser_pi1_marker.php' ] )
{
	include_once($TYPO3_CONF_VARS[ TYPO3_MODE ][ 'XCLASS' ][ 'ext/browser/Classes/Plugins/Pi1/class.tx_browser_pi1_marker.php' ]);
}