<?php

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2010-2019 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * The class tx_browser_pi1_socialmedia bundles methods for social media for the extension browser
 *
 * @author    Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package    TYPO3
 * @subpackage  browser
 *
 * @version 9.2.7
 * @version 3.0.0
 */
class tx_browser_pi1_socialmedia
{

	//////////////////////////////////////////////////////
	//
  // Variables set by the pObj (by class.tx_browser_pi1.php)

	var $conf = false;
	// [Array] The current TypoScript configuration array
	var $mode = false;
	// [Integer] The current mode (from modeselector)
	var $view = false;
	// [String] 'list' or 'single': The current view
	var $conf_view = false;
	// [Array] The TypoScript configuration array of the current view
	var $conf_path = false;
	// [String] TypoScript path to the current view. I.e. views.single.1
	// Variables set by the pObj (by class.tx_browser_pi1.php)
	/////////////////////////////////////////////////
	//
  // Vars set by methods in the current class

	var $str_htmlBookmarks_list = false;
	// [string] html bookmark code for list views. It contains the marker ###URL### and ###TITLE###
	var $str_htmlBookmarks_single = false;
	// [string] html bookmark code for single views. It contains the marker ###URL### and ###TITLE###
//  var $arr_default_stdWrapItem   = false;
//  // [array] TypoScript configuration array for a stdWrap of a bookmark (one item). Elements: ['10'] and ['10.']
//  var $arr_default_stdWrapItems  = false;
//  // [array] TypoScript configuration array for a stdWrap of bookmarks (items list). Elements: ['10'] and ['10.']
	var $str_default_htmlItem = false;
	// [string] Default HTML code for a bookmark (one item).
	//          This marker will be replaced: ###BOOKMARK_NAME###, ###BOOKMARK_URL###, ###BOOKMARK_IMAGE###, ###BOOKMARK_IMAGESIZE###
	//          This marker won't be replaced: ###URL### and ###TITLE###
	var $str_default_htmlItems = false;

	// [string] Default HTML code for bookmarks (items list).
	// Vars set by methods in the current class

	/**
	 * Constructor. The method initiate the parent object
	 *
	 * @param	object		The parent object
	 * @return	void
	 */
	function __construct( $parentObj )
	{
		$this->pObj = $parentObj;
	}

	/*	 * *********************************************
	 *
	 * Bookmarks
	 *
	 * ******************************************** */

	/**
	 * get_htmlBookmarks: Set the HTML code for bookmarks in a list or single view.
	 *                    Result will be written to the globals $str_htmlbookmarks_list or $str_htmlbookmarks_single.
	 *                    Method is available once only.
	 *
	 * @param	array		$elements: The SQL result of the current row
	 * @param	string		$key: The name current table.field
	 * @param	boolean		$bool_defaultTemplate: TRUE if we have a default HTML template (with ###ITEM### marker)
	 * @return	string		$str_items or FALSE. $str_items are the rendered bookmarks
	 * @version 8.0.38
	 * @since 3.0.0
	 */
	public function get_htmlBookmarks( $elements, $key, $bool_defaultTemplate )
	{
		static $int_countThisMethod = 1;
		static $int_currPlugin_uid = 0;

		// #9596: New plugin: Method is called the first time
		if( $int_currPlugin_uid != $this->pObj->cObj->data[ 'uid' ] )
		{
			$int_countThisMethod = 1;
			$int_currPlugin_uid = $this->pObj->cObj->data[ 'uid' ];
		}

		// #i0336, 170409, dwildt, ~
		if( !$this->init( $bool_defaultTemplate, $key ) )
		{
			return;
		}

		// Individual template (without ###ITEM### or ###VALUE###)
		// DRS - Development Reporting System
		// HTML bookmark template
		//        This marker will be replaced: ###BOOKMARK_NAME###, ###BOOKMARK_URL###, ###BOOKMARK_IMAGE###, ###BOOKMARK_IMAGESIZE###
		//          This marker won't be replaced: ###URL### and ###TITLE###
		$str_items = $this->htmlBookmarks();

		// #i0336, 170409, dwildt, ~
		if( empty( $str_items ) )
		{
			return;
		}

		$str_items = $this->replaceParams( $str_items, $elements );
		return $str_items;
	}

	/*	 * *********************************************
	 *
	 * Initial methods
	 *
	 * ******************************************** */

	/**
	 * bookmarksFromFlexform(): Get bookmarks from the flexform
	 *
	 * @access private
	 * @version 8.0.38
	 * @since 3.0.0
	 * @return	array		$arr_currBookmarks
	 */
	private function bookmarksFromFlexform()
	{
		switch( true )
		{
			case($this->view == 'list'):
				$csv_bookmarks = $this->pObj->objFlexform->strCsv_socialmedia_bookmarks_list;
				break;
			case($this->view == 'single'):
				$csv_bookmarks = $this->pObj->objFlexform->strCsv_socialmedia_bookmarks_single;
				break;
			default:
				$prompt = 'Undefined parameter in switch at ' . __METHOD__ . ' #' . __LINE__;
				die( $prompt );
				break;
		}

		if( $csv_bookmarks )
		{
			$arr_currBookmarks = explode( ',', $csv_bookmarks );
			if( !$this->pObj->b_drs_socialmedia )
			{
				return $arr_currBookmarks;
			}
			$prompt = 'Bookmarks for a the current view: ' . $csv_bookmarks;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 0 );
			return $arr_currBookmarks;
		}

		if( !$this->pObj->b_drs_socialmedia )
		{
			return false;
		}

		$prompt = 'There isn\'t any bookmark item selected in the plugin.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 0 );
		return false;
	}

	/**
	 * htmlBookmark: 
	 *
	 * @access private
	 * @version 8.0.38
	 * @since 8.0.38
	 * @return	string
	 */
	private function htmlBookmark( $arr_tsBookmarks, $str_currBookmark )
	{
		// Get the name
		$markerArray[ '###BOOKMARK_NAME###' ] = $arr_tsBookmarks[ $str_currBookmark . '.' ][ 'name' ];
		if( !$markerArray[ '###BOOKMARK_NAME###' ] )
		{
			$markerArray[ '###BOOKMARK_NAME###' ] = $str_currBookmark;
		}
		// Get the name
		// Get the url
		$markerArray[ '###BOOKMARK_URL###' ] = $arr_tsBookmarks[ $str_currBookmark . '.' ][ 'url' ];
		// #i0336, 170409, dwildt, ~
		$html = $arr_tsBookmarks[ $str_currBookmark . '.' ][ 'html' ];
		if( empty( $html ) )
		{
			$html = '<span style="display:none">ERROR. See ' . __METHOD__ . ' #' . __LINE__ . '</span>';
		}
		$markerArray[ '###HTML###' ] = $html;
		// Get the image path
		$str_imagePath = $arr_tsBookmarks[ $str_currBookmark . '.' ][ 'image' ];
		// Change EXT:browser ... to typo3/typo3conf/ext/browser ...
		$str_imagePathRel = $this->pObj->objZz->get_pathWoEXT( $str_imagePath );
		$str_imagePathAbs = \TYPO3\CMS\Core\Utility\GeneralUtility::getFileAbsFileName( $str_imagePath );
		// Get the image size
		$arr_imageSize = getimagesize( $str_imagePathAbs );
		// Image size isn't proper
		if( !is_array( $arr_imageSize ) )
		{
			$markerArray[ '###BOOKMARK_IMAGESIZE###' ] = 'witdh="16" height="16"';
			// DRS - Development Reporting System
			if( $this->pObj->b_drs_warn )
			{
				$prompt = 'Image size isn\'t proper: socialmedia.bookmarks.items.' . $str_currBookmark;
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SOCIALMEDIA] ' . $prompt . ' = \'' . $str_imagePathAbs . '\'', $this->pObj->extKey, 2 );
			}
			// DRS - Development Reporting System
		}
		// Image size isn't proper
		// Image size is proper
		if( is_array( $arr_imageSize ) )
		{
			$markerArray[ '###BOOKMARK_IMAGESIZE###' ] = $arr_imageSize[ 3 ];
		}
		// Image size is proper
		$markerArray[ '###BOOKMARK_IMAGE###' ] = $str_imagePathRel;
		// Default HTML code for the item
		$str_item = $this->str_default_htmlItem;
		// Item has its own stdWrap
		$arr_lConf = $arr_tsBookmarks[ $str_currBookmark . '.' ][ 'stdWrap.' ];
		if( is_array( $arr_lConf ) && count( $arr_lConf ) > 0 )
		{
			$str_item = $this->pObj->local_cObj->stdWrap( $arr_lConf[ 'value' ], $arr_lConf );
		}
		// Item has its own stdWrap
		$str_item = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $str_item, $markerArray );
		//var_dump( __METHOD__, __LINE__, $str_item );
		return $str_item;
		// Item array
	}

	/**
	 * htmlBookmarks: Set the HTML code (template) for bookmarks in a list or single view.
	 *                     Result will be written to the globals $str_htmlbookmarks_list or $str_htmlbookmarks_single.
	 *                     Method is available once only.
	 *                     This marker will be replaced: ###BOOKMARK_NAME###, ###BOOKMARK_URL###, ###BOOKMARK_IMAGE###, ###BOOKMARK_IMAGESIZE###
	 *                     This marker won't be replaced: ###URL### and ###TITLE###
	 *
	 * @access private
	 * @version 8.0.38
	 * @since 3.0.0
	 * @return	string		$str_items or false. $str_items is an html template code with ###URL### and ###TITLE###
	 */
	private function htmlBookmarks()
	{
		static $int_countThisMethod = 1;
		static $int_currPlugin_uid = 0; // #9596
		static $str_items = '';

		// #9596: New plugin: Method is called the first time
		if( $int_currPlugin_uid != $this->pObj->cObj->data[ 'uid' ] )
		{
			$int_countThisMethod = 1;
			$int_currPlugin_uid = $this->pObj->cObj->data[ 'uid' ];
			$str_items = '';
		}

		if( !empty( $str_items ) )
		{
			return $str_items;
		}

		$arr_currBookmarks = $this->bookmarksFromFlexform();
		if( empty( $arr_currBookmarks ) )
		{
			$int_countThisMethod = $int_countThisMethod + 1;
			return false;
		}

		$arr_tsBookmarks = $this->tsBookmarks();
		if( empty( $arr_tsBookmarks ) )
		{
			$int_countThisMethod = $int_countThisMethod + 1;
			return false;
		}

		// Init the default stdWraps for an item (one bookmark) and the item list (bookmarks)
		$this->htmlBookmarksDefaultStdWrap();

		// Loop through the bookmarks, which are selected in the flexform plugin
		$arr_item = false;
		foreach( ( array ) $arr_currBookmarks as $str_currBookmark )
		{
			if( !empty( $arr_tsBookmarks[ $str_currBookmark . '.' ] ) )
			{
				$arr_item[] = $this->htmlBookmark( $arr_tsBookmarks, $str_currBookmark );
				continue;
			}

			if( !$this->pObj->b_drs_warn )
			{
				continue;
			}

			$prompt = 'Current Bookmark isn\'t a valid TypoScript array: socialmedia.bookmarks.items.' . $str_currBookmark;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 2 );
			$prompt = 'Please maintain your TypoScript or your plugin.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 1 );
		}

		// Loop through the bookmarks, which are selected in the flexform plugin
		// Wrap all bookmarks
		$markerArray[ '###BOOKMARK_ITEMS###' ] = implode( '', ( array ) $arr_item );

		// Default HTML code for the item
		$str_items = $this->str_default_htmlItems;
		$str_items = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $str_items, $markerArray );
		// Wrap all bookmarks
		//if(\TYPO3\CMS\Core\Utility\GeneralUtility::_GP('dev')) var_dump('socialmedia 379', $markerArray, $str_items);
		// Set current bookmars for the list view
		if( $this->view == 'list' )
		{
			$this->str_htmlbookmarks_list = $str_items;
		}
		// Set current bookmars for the list view
		// Set current bookmars for the single view
		if( $this->view == 'single' )
		{
			$this->str_htmlbookmarks_single = $str_items;
		}
		// Set current bookmars for the single view
		// Next time isn't the first time
// 100912
		$int_countThisMethod = $int_countThisMethod + 1;

		// Return the html template code with ###URL### and ###TITLE###
		return $str_items;
	}

	/**
	 * init(): 
	 * 
	 * @param	boolean		$bool_defaultTemplate: TRUE if we have a default HTML template (with ###ITEM### marker)
	 * @param	string		$key: The name current table.field
	 * @return	boolean
	 * @access private
	 * @version 8.0.38
	 * @since 8.0.38
	 */
	private function init( $bool_defaultTemplate, $key )
	{
		if( !$this->initEnabled() )
		{
			return false;
		}
		if( !$this->initTemplate( $bool_defaultTemplate, $key ) )
		{
			return false;
		}


		return true;
	}

	/**
	 * initEnabled(): 
	 * 
	 * @return	boolean
	 * @access private
	 * @version 8.0.38
	 * @since 8.0.38
	 */
	private function initEnabled()
	{
		static $int_countThisMethod = 1;

		$str_bookmarks_enabled = $this->pObj->objFlexform->str_socialmedia_bookmarks_enabled;
		switch( true )
		{
			case($str_bookmarks_enabled == 'disabled'):
				$returnTrue = false;
				break;
			case(!$str_bookmarks_enabled):
				$returnTrue = false;
				break;
			default:
				$returnTrue = true;
				break;
		}
		if( $returnTrue )
		{
			return true;
		}

		if( $int_countThisMethod > 1 )
		{
			return false;
		}

		if( $this->pObj->b_drs_socialmedia )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] RETURN. Social Bookmarks are disabled.', $this->pObj->extKey, 0 );
		}

		$int_countThisMethod = $int_countThisMethod + 1;

		return false;
	}

	/**
	 * initTemplate(): 
	 * 
	 * @param	boolean		$bool_defaultTemplate: TRUE if we have a default HTML template (with ###ITEM### marker)
	 * @param	string		$key: The name current table.field
	 * @return	boolean
	 * @access private
	 * @version 8.0.68
	 * @since 8.0.38
	 */
	private function initTemplate( $bool_defaultTemplate, $key )
	{
		static $int_countThisMethod = 1;

		// Default HTML template with an ###ITEM### marker or a ###VALUE### marker
		if( !$bool_defaultTemplate )
		{
			$this->initTemplateDRSNotDefault( $int_countThisMethod );
			$int_countThisMethod = $int_countThisMethod + 1;
			return true;
		}

		// RETURN false if current key is the table.field for site
		switch( true )
		{
			case($this->view == 'list'):
				$str_currKey_forSite = $this->pObj->objFlexform->str_socialmedia_bookmarks_tableFieldSite_list;
				break;
			case($this->view == 'single'):
				$str_currKey_forSite = $this->pObj->objFlexform->str_socialmedia_bookmarks_tableFieldSite_single;
				break;
			default:
				$prompt = 'Undefined parameter in switch at ' . __METHOD__ . ' #' . __LINE__;
				die( $prompt );
				break;
		}

		if( $str_currKey_forSite == $key )
		{
			if( $int_countThisMethod > 1 )
			{
				return true;
			}
			if( $this->pObj->b_drs_socialmedia )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] HTML template with ###ITEM### marker or ###VALUE### marker. ' .
								'Current key is table.field for site (' . $str_currKey_forSite . ').', $this->pObj->extKey, 0 );
			}
			$int_countThisMethod = $int_countThisMethod + 1;
			return true;
		}

		if( $int_countThisMethod > 1 )
		{
			return false;
		}

		if( $this->pObj->b_drs_socialmedia )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] RETURN. HTML template with ###ITEM### marker or ###VALUE### marker ' .
							'Current key (' . $key . ') isn\'t table.field for site (' . $str_currKey_forSite . ').', $this->pObj->extKey, 0 );
		}

		$int_countThisMethod = $int_countThisMethod + 1;
		return false;
	}

	/**
	 * initTemplateDRSNotDefault(): 
	 * 
	 * @param	$int_countThisMethod
	 * @return	void
	 * @access private
	 * @version 8.0.38
	 * @since 8.0.38
	 */
	private function initTemplateDRSNotDefault( $int_countThisMethod )
	{
		if( $int_countThisMethod > 1 )
		{
			return;
		}
		if( $this->pObj->b_drs_socialmedia )
		{
			$prompt = '[INFO/SOCIALMEDIA] HTML template without ###ITEM### marker or ###VALUE### marker. ';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( $prompt, $this->pObj->extKey, 0 );
		}
	}

	/**
	 * htmlBookmarksDefaultStdWrap: Init the default stdWraps for items (all bookmarks) and item (one bookmark)
	 *
	 * @return	boolean		false
	 */
	private function htmlBookmarksDefaultStdWrap()
	{
		static $bool_firsttime = true;
		static $int_currPlugin_uid = 0; // #9596
		// #9596: New plugin: Method is called the first time
		if( $int_currPlugin_uid != $this->pObj->cObj->data[ 'uid' ] )
		{
			$bool_firsttime = true;
			$int_currPlugin_uid = $this->pObj->cObj->data[ 'uid' ];
		}
		// #9596: New plugin: Method is called the first time
		// RETURN if HTML code is set for bookmars in a list view
		if( !$bool_firsttime )
		{
			return false;
		}
		// RETURN if HTML code is set for bookmars in a list view
		// Get the local stdWrap for an item
		$arr_default_stdWrapItem = $this->conf_view[ 'flexform.' ][ 'socialmedia.' ][ 'socialbookmarks.' ][ 'wraps.' ][ 'stdWrap_item.' ];
		if( !is_array( $arr_default_stdWrapItem ) || count( $arr_default_stdWrapItem ) < 1 )
		{
			// DRS - Development Reporting System
			if( $this->pObj->b_drs_socialmedia )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] There is no local default stdWrap for a bookmark item!', $this->pObj->extKey, 0 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] Try the global one ...', $this->pObj->extKey, 0 );
			}
			// DRS - Development Reporting System
		}
		// Get the local stdWrap for an item
		// Get the global stdWrap for an item
		if( !is_array( $arr_default_stdWrapItem ) || count( $arr_default_stdWrapItem ) < 1 )
		{
			$arr_default_stdWrapItem = $this->conf[ 'flexform.' ][ 'socialmedia.' ][ 'socialbookmarks.' ][ 'wraps.' ][ 'stdWrap_item.' ];
			// WARN if there is no global array for a stdWrap
			if( !is_array( $arr_default_stdWrapItem ) || count( $arr_default_stdWrapItem ) < 1 )
			{
				$this->str_default_htmlItem = false;
				// DRS - Development Reporting System
				if( $this->pObj->b_drs_warn )
				{
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SOCIALMEDIA] There is no default stdWrap for a bookmark item!', $this->pObj->extKey, 2 );
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/SOCIALMEDIA] Please configure socialmedia.wraps.stdWrap_item', $this->pObj->extKey, 1 );
				}
				// DRS - Development Reporting System
			}
			// WARN if there is no global array for a stdWrap
		}
		// Get the global stdWrap for an item
		// TS config array for the item is proper
		if( is_array( $arr_default_stdWrapItem ) || count( $arr_default_stdWrapItem ) > 0 )
		{
			// Default HTML code of a bookmark (one item)
			$this->str_default_htmlItem = $this->pObj->local_cObj->stdWrap( $arr_default_stdWrapItem[ 'value' ], $arr_default_stdWrapItem );
		}
		// TS config array for the item is proper
		// TS config array result for the item isn't proper
		if( !$this->str_default_htmlItem )
		{
			// DRS - Development Reporting System
			if( $this->pObj->b_drs_error )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SOCIALMEDIA] There is no result for socialmedia.wraps.stdWrap_item!', $this->pObj->extKey, 3 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] tsConfig: ' . $arr_default_stdWrapItem, $this->pObj->extKey, 1 );
			}
			// DRS - Development Reporting System
		}
		// TS config array result for the item isn't proper
		// Get the local array for the stdWrap for items (bookmark list)
		$arr_default_stdWrapItems = $this->conf_view[ 'flexform.' ][ 'socialmedia.' ][ 'socialbookmarks.' ][ 'wraps.' ][ 'stdWrap_items.' ];
		if( !is_array( $arr_default_stdWrapItems ) || count( $arr_default_stdWrapItems ) < 1 )
		{
			// DRS - Development Reporting System
			if( $this->pObj->b_drs_socialmedia )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] There is no local default stdWrap for the bookmarks (items list)!', $this->pObj->extKey, 0 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] Try the global one ...', $this->pObj->extKey, 0 );
			}
			// DRS - Development Reporting System
		}
		// Get the local array for the stdWrap for items (bookmark list)
		// Get the global array for the stdWrap for items (bookmark list)
		if( !is_array( $arr_default_stdWrapItems ) || count( $arr_default_stdWrapItems ) < 1 )
		{
			$arr_default_stdWrapItems = $this->conf[ 'flexform.' ][ 'socialmedia.' ][ 'socialbookmarks.' ][ 'wraps.' ][ 'stdWrap_items.' ];
			if( !is_array( $arr_default_stdWrapItems ) || count( $arr_default_stdWrapItems ) < 1 )
			{
				$this->str_default_htmlItems = false;
				// DRS - Development Reporting System
				if( $this->pObj->b_drs_warn )
				{
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SOCIALMEDIA] There is no global default stdWrap for the bookmarks (items list)!', $this->pObj->extKey, 2 );
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/SOCIALMEDIA] Please configure socialmedia.wraps.stdWrap_items', $this->pObj->extKey, 1 );
				}
				// DRS - Development Reporting System
			}
		}
		// Get the global array for the stdWrap for items (bookmark list)
		// TS config array for the items is proper
		if( is_array( $arr_default_stdWrapItems ) || count( $arr_default_stdWrapItems ) > 0 )
		{
			// Default HTML code of a bookmark (one item)
			$this->str_default_htmlItems = $this->pObj->local_cObj->stdWrap( $arr_default_stdWrapItems[ 'value' ], $arr_default_stdWrapItems );
		}
		// TS config array for the item is proper
		// TS config array result for the item isn't proper
		if( !$this->str_default_htmlItems )
		{
			// DRS - Development Reporting System
			if( $this->pObj->b_drs_error )
			{
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SOCIALMEDIA] There is no result for socialmedia.wraps.stdWrap_items!', $this->pObj->extKey, 3 );
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] tsConfig: ' . $arr_default_stdWrapItems, $this->pObj->extKey, 1 );
			}
			// DRS - Development Reporting System
		}
		// TS config array result for the item isn't proper
		// Next time isn't the first time!
// 100912
		$bool_firsttime = false;
	}

	/**
	 * replaceParams(): 
	 * 
	 * @param	string		$items
	 * @param	array		$elements: The SQL result of the current row
	 * @return	string
	 * @access private
	 * @internal #t2258
	 * @version 9.2.7
	 * @since 8.0.38
	 */
	private function replaceParams( $items, $elements )
	{
		$markerArray = false;

		// #t2258, 180427, dwildt, 3+-/2++
		$cObj_name = $this->pObj->conf[ 'displayList.' ][ 'master_templates.' ][ 'tableFields.' ][ 'typolinksAbsoluteURL' ];
		$cObj_conf = $this->pObj->conf[ 'displayList.' ][ 'master_templates.' ][ 'tableFields.' ][ 'typolinksAbsoluteURL.' ];
		$url = $this->pObj->cObj->cObjGetSingle( $cObj_name, $cObj_conf );

		$markerArray[ '###URL###' ] = $url;
		$markerArray[ '###TITLE###' ] = $this->replaceParamsTitle( $elements );

		// Replace ###URL### and ###TITLE###
		$items = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $items, $markerArray );

		return $items;
	}

//	/**
//	 * replaceParamsShowUid(): 
//	 * 
//	 * @param	array		$elements: The SQL result of the current row
//	 * @return	void
//	 * @access private
//	 * @version 9.2.7
//	 * @since 8.0.38
//	 */
//	private function replaceParamsShowUid( $elements )
//	{
//		// Alias for showUid?
//		// 13930, 110320, dwildt
//		$bool_rmShowUid = false;
//		$bool_setShowUid = false;
//		$bool_rmAliasUid = false;
//
//		// No alias for showUid
//		if( empty( $this->pObj->piVar_alias_showUid ) )
//		{
//			// SET piVar showUid
//			if( !isset( $this->pObj->piVars[ 'showUid' ] ) )
//			{
//				$bool_rmShowUid = true;
//				$uidField = $this->pObj->arrLocalTable[ 'uid' ];
//				$int_recUid = $elements[ $uidField ];
//				$this->pObj->piVars[ 'showUid' ] = $int_recUid;
//			}
//			// SET piVar showUid
//		}
//		// No alias for showUid
//		// Alias for showUid
//		// 13930, 110320, dwildt
//		if( !empty( $this->pObj->piVar_alias_showUid ) )
//		{
//			if( isset( $this->pObj->piVars[ 'showUid' ] ) )
//			{
//				$bool_setShowUid = true;
//				$int_showUid = $this->pObj->piVars[ 'showUid' ];
//				unset( $this->pObj->piVars[ 'showUid' ] );
//			}
//			// SET piVar alias (for showUid)
//			if( !isset( $this->pObj->piVars[ $this->pObj->piVar_alias_showUid ] ) )
//			{
//				$bool_rmAliasUid = true;
//				$uidField = $this->pObj->arrLocalTable[ 'uid' ];
//				$int_recUid = $elements[ $uidField ];
//				$this->pObj->piVars[ $this->pObj->piVar_alias_showUid ] = $int_recUid;
//			}
//			// SET piVar alias (for showUid)
//		}
//		// Alias for showUid
//		// Set additional Parameter and the cHash
//		// Remove the piVar for the record uid
//		if( $bool_rmShowUid )
//		{
//			unset( $this->pObj->piVars[ 'showUid' ] );
//		}
//		if( $bool_rmAliasUid )
//		{
//			unset( $this->pObj->piVars[ $this->pObj->piVar_alias_showUid ] );
//		}
//		// Remove the piVar for the record uid
//		// Reset showUid
//		// 13930, 110320, dwildt
//		if( $bool_setShowUid )
//		{
//			$this->pObj->piVars[ 'showUid' ] = $int_showUid;
//		}
//	}
//
//	/**
//	 * replaceParamsSinglePid(): 
//	 * 
//	 * @return	string
//	 * @access private
//	 * @internal #t2258
//	 * @version 9.2.7
//	 * @since 8.0.38
//	 */
//	private function replaceParamsSinglePid()
//	{
//		switch( TRUE )
//		{
//			case( $this->view == 'list' ):
//				$singlePid = $this->pObj->objZz->get_singlePid_for_listview();
//				break;
//			case( $this->view == 'single' ):
//			default:
//				$singlePid = $GLOBALS[ 'TSFE' ]->id;
//				break;
//		}
//		return $singlePid;
//	}

	/**
	 * replaceParamsTitle(): 
	 * 
	 * @return	string
	 * @access private
	 * @internal #t2258
	 * @version 9.2.7
	 * @since 8.0.38
	 */
	private function replaceParamsTitle( $elements )
	{
		switch( TRUE )
		{
			case( $this->view == 'list' ):
				$field = $this->pObj->objFlexform->str_socialmedia_bookmarks_tableFieldTitle_list;
				break;
			case( $this->view == 'single' ):
			default:
				$field = $this->pObj->objFlexform->str_socialmedia_bookmarks_tableFieldTitle_single;
				break;
		}

		switch( TRUE )
		{
			case( isset( $elements[ $field ] ) ):
				$value = $elements[ $field ];
				break;
			case(!isset( $elements[ $field ] ) ):
			default:
				$value = $this->pObj->objTemplate->arr_curr_value[ $field ];
				break;
		}

		return rawurlencode( $value );
	}

//	/**
//	 * replaceParams(): 
//	 * 
//	 * @param	string		$str_items
//	 * @param	array		$elements: The SQL result of the current row
//	 * @return	string
//	 * @access private
//	 * @version 9.2.7
//	 * @since 8.0.38
//	 */
//	private function replaceParamsBAK( $str_items, $elements )
//	{
//		$singlePid = $this->replaceParamsSinglePid();
//
//		$this->replaceParamsShowUid( $elements );
////		// Get the uid of the page with the single view
////		// Alias for showUid?
////		// 13930, 110320, dwildt
////		$bool_rmShowUid = false;
////		$bool_setShowUid = false;
////		$bool_rmAliasUid = false;
////
////		// No alias for showUid
////		if( empty( $this->pObj->piVar_alias_showUid ) )
////		{
////			// SET piVar showUid
////			if( !isset( $this->pObj->piVars[ 'showUid' ] ) )
////			{
////				$bool_rmShowUid = true;
////				$uidField = $this->pObj->arrLocalTable[ 'uid' ];
////				$int_recUid = $elements[ $uidField ];
////				$this->pObj->piVars[ 'showUid' ] = $int_recUid;
////			}
////			// SET piVar showUid
////		}
////		// No alias for showUid
////		// Alias for showUid
////		// 13930, 110320, dwildt
////		if( !empty( $this->pObj->piVar_alias_showUid ) )
////		{
////			if( isset( $this->pObj->piVars[ 'showUid' ] ) )
////			{
////				$bool_setShowUid = true;
////				$int_showUid = $this->pObj->piVars[ 'showUid' ];
////				unset( $this->pObj->piVars[ 'showUid' ] );
////			}
////			// SET piVar alias (for showUid)
////			if( !isset( $this->pObj->piVars[ $this->pObj->piVar_alias_showUid ] ) )
////			{
////				$bool_rmAliasUid = true;
////				$uidField = $this->pObj->arrLocalTable[ 'uid' ];
////				$int_recUid = $elements[ $uidField ];
////				$this->pObj->piVars[ $this->pObj->piVar_alias_showUid ] = $int_recUid;
////			}
////			// SET piVar alias (for showUid)
////		}
////		// Alias for showUid
//		// Set additional Parameter and the cHash
//		$additionalParams = null;
//		foreach( ( array ) $this->pObj->piVars as $paramKey => $paramValue )
//		{
//			if( !empty( $paramValue ) )
//			{
//				$additionalParams .= '&' . $this->pObj->prefixId . '[' . $paramKey . ']=' . $paramValue;
//			}
//		}
////		// Set additional Parameter and the cHash
////		// Remove the piVar for the record uid
////		if( $bool_rmShowUid )
////		{
////			unset( $this->pObj->piVars[ 'showUid' ] );
////		}
////		if( $bool_rmAliasUid )
////		{
////			unset( $this->pObj->piVars[ $this->pObj->piVar_alias_showUid ] );
////		}
////		// Remove the piVar for the record uid
////		// Reset showUid
////		// 13930, 110320, dwildt
////		if( $bool_setShowUid )
////		{
////			$this->pObj->piVars[ 'showUid' ] = $int_showUid;
////		}
//		// #t2258, 180427, dwildt, 3+-/2++
//		$cObj_name = $this->pObj->conf[ 'displayList.' ][ 'master_templates.' ][ 'tableFields.' ][ 'typolinksAbsoluteURL' ];
//		$cObj_conf = $this->pObj->conf[ 'displayList.' ][ 'master_templates.' ][ 'tableFields.' ][ 'typolinksAbsoluteURL.' ];
//		$url = $this->pObj->cObj->cObjGetSingle( $cObj_name, $cObj_conf );
//		$markerArray = false;
//		$markerArray[ '###URL###' ] = $url;
////var_dump( __METHOD__, __LINE__, $url );
////die();
//		// #t2258, 180427, dwildt, -
////		// Reset showUid
////		// Set the TypoScript configuration array
////		// #t2257, 180425, dwildt, :TODO:
////		//   Replace PHP generated typolink by 
////		$lConfCObj = null;
////		$lConfCObj[ 'typolink.' ][ 'parameter' ] = $singlePid;
////		// #54983, 140624, dwildt, 2-
////		//$cHash_calc = $this->pObj->objZz->get_cHash('&id='.$singlePid.$additionalParams);
////		//$lConfCObj['typolink.']['additionalParams']  = $additionalParams.'&cHash='.$cHash_calc;
////		// #54983, 140624, dwildt, 1+
////		$lConfCObj[ 'typolink.' ][ 'additionalParams' ] = $additionalParams;
////		$lConfCObj[ 'typolink.' ][ 'returnLast' ] = 'url';
////		// #54983, 140624, dwildt, 1+
////		$lConfCObj[ 'typolink.' ][ 'useCacheHash' ] = '1';
////		// #t2258, 180427, dwildt, 1+
////		$lConfCObj[ 'typolink.' ][ 'forceAbsoluteUrl' ] = '1';
////		// Set the TypoScript configuration array
////		// Set the URL (wrap the Link)
////		$str_relUrl = $this->pObj->local_cObj->stdWrap( '#', $lConfCObj );
////		// #t2258, 180427, dwildt, 3-/2+
//////		$str_absUrl = $this->pObj->objZz->get_absUrl( $str_relUrl );
//////		$markerArray = false;
//////		$markerArray[ '###URL###' ] = $str_absUrl;
////		$markerArray = false;
////		$markerArray[ '###URL###' ] = $str_relUrl;
//		// Set the URL (wrap the Link)
//		// Set the title property
//		if( isset( $elements[ $str_tableFieldTitle ] ) )
//		{
//			$str_title = $elements[ $str_tableFieldTitle ];
//		}
//		if( !isset( $elements[ $str_tableFieldTitle ] ) )
//		{
//			$str_title = $this->pObj->objTemplate->arr_curr_value[ $str_tableFieldTitle ];
//		}
//		$str_title = rawurlencode( $str_title );
//		$markerArray[ '###TITLE###' ] = $str_title;
//		// Set the title property
//		// Replace ###URL### and ###TITLE###
//		$str_items = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $str_items, $markerArray );
//		// Replace ###URL### and ###TITLE###
//		// Set the link to the single view
//		// Return the result
//		$int_countThisMethod = $int_countThisMethod + 1;
//
//		return $str_items;
//	}

	/**
	 * tsBookmarks(): Get typoScript configuration
	 *
	 * @access private
	 * @version 8.0.38
	 * @since 3.0.0
	 * @return	array		$arr_tsBookmarks
	 */
	private function tsBookmarks()
	{
		// Get the TypoScript local bookmark array
		$arr_tsBookmarks = $this->conf_view[ 'flexform.' ][ 'socialmedia.' ][ 'socialbookmarks.' ][ 'bookmarks.' ][ 'items.' ];
		if( !empty( $arr_tsBookmarks ) )
		{
			return $arr_tsBookmarks;
		}

		if( $this->pObj->b_drs_socialmedia )
		{
			$prompt = 'There is no local bookmark array in TypoScript.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 0 );
			$prompt = 'Try to get the global one ...';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 0 );
		}

		// Get the TypoScript global bookmark array
		$arr_tsBookmarks = $this->conf[ 'flexform.' ][ 'socialmedia.' ][ 'socialbookmarks.' ][ 'bookmarks.' ][ 'items.' ];
		if( !empty( $arr_tsBookmarks ) )
		{
			return $arr_tsBookmarks;
		}

		if( $this->pObj->b_drs_warn )
		{
			$prompt = 'There is no global bookmark array in TypoScript!';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 2 );
			$prompt = 'Please configure socialmedia.bookmarks.items.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/SOCIALMEDIA] ' . $prompt, $this->pObj->extKey, 1 );
		}

		return false;
	}

}

if( defined( 'TYPO3_MODE' ) && $TYPO3_CONF_VARS[ TYPO3_MODE ][ 'XCLASS' ][ 'ext/browser/Classes/Plugins/Pi1/class.tx_browser_pi1_socialmedia.php' ] )
{
	include_once($TYPO3_CONF_VARS[ TYPO3_MODE ][ 'XCLASS' ][ 'ext/browser/Classes/Plugins/Pi1/class.tx_browser_pi1_socialmedia.php' ]);
}